/* MINIX IDE driver for A1200
 * The code uses parts of the existing MINIX Amiga driver plus code from the freely available MINDRVR.C
 * minimalist ATA driver by Hale Landis.
 * Other sources of information for the inner workings of the IDE on A1200:
 * AROS, EmuTOS, Linux, Gayle specifications document
 *
 * The driver is composed of three files:
 * amatai.c (this file)
 * mindrvr.c (an abstract of the full mindrvr.c ATA driver
 * mindrvr.h (an abstract of the full mindrvr.h header file
 *
 * A. Eleuteri 12/2020
*/
 
#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include <minix/partition.h>
#include <time.h>
#include "proc.h"
#include "proto.h"
#include "string.h"
#include "amhardware.h"
#include "mindrvr.h"


PUBLIC int has_gayle;
PRIVATE message w_mess;
PRIVATE message clock_mess;
PRIVATE message dummy_mess;
PRIVATE char *xfer_buf;
FORWARD char *BtoCstr();


/* Gayle registers */
#define GAYLE_BASE 			((void *)0xde1000)
#define GAYLE_ID 			*(unsigned char*)0xde1000
#define FAT_GARY_TIMEOUT 	*(unsigned char*)0xde0000
#define INTENA_MIRROR 		(unsigned short*)0xde109a

#define GAYLE_IRQ_1200  	(unsigned char *)0xda9000 /* Interrupt request */
#define GAYLE_INT_1200  	(unsigned char *)0xdaa000 /* Interrupt enable */

#define GAYLE_IRQ_IDE		0x80
#define GAYLE_INT_IDE		0x80
#define LOCK				*INTENA=0x4000
#define UNLOCK				*INTENA=0xC000

/*
 *  Miscellaneous
 */

#define FORCE_READ(x)   (void)(x)     /* Read a volatile hardware register */

#define SECTOR_SIZE	512		/* physical sector size in bytes */
/*
 * maximum number of sectors per physical i/o.  this MUST not exceed 256,
 * at least for LBA28-style commands.  the best performance is obtained
 * if this is a multiple of the sectors-per-interrupt value supported by
 * the drive(s) in multiple mode.
 */

#define MAXSECS_PER_IO  32

#define MaxTransfer	(MAXSECS_PER_IO * SECTOR_SIZE)

#define ID_RDSK		0x5244534b
#define ID_PART		0x50415254

/*PRIVATE long block[SECTOR_SIZE / sizeof(long)];
*/

struct part {
	char *p_name;
	u_long p_start;			/* first byte (mult of SECTOR_SIZE) */
	u_long p_size;			/* size in bytes */
};

/* 
IMPLEMENTATION NOTE:

4 bits are used to store partition information
- bits 0-2 for the AmigaDOS partition (0=whole disk)
- bit 3 for the ATA unit (0=master, 1=slave)

I assume that only one ATA interface is on the system (i.e. no IDE-doubler)
This makes the driver a lot simpler, as the MINDRV code assumes a single interface.

0 = 0000 -> part[0]   	master unit, whole disk
1 = 0001 -> part[1]		master unit, partition 1
2 = 0010 -> part[2]		master unit, partition 2
			...
7 = 0111 -> part[7]		master unit, partition 7

8 = 1000 -> part[8]   	slave unit, whole disk
9 = 1001 -> part[9]		slave unit, partition 1
10= 1010 -> part[10]	slave unit, partition 2
			...
15= 1111 -> part[15]	slave unit, partition 7

*/

PRIVATE struct part part[16];		/* one for each minor dev number */


/* Detect A600 / A1200 Gayle chip.
 * Freely inspired from AROS ReadGayle().
 * Adapted from EmuTOS implementation.
 */
PRIVATE void detect_gayle()
{
	u_short save_intena;
	u_char gayle_id;
	int i;

	has_gayle = 0;

	/* Check if 0xde1000 is a mirror of 0xdff000 custom chips */
	save_intena = *INTENAR;
	*INTENA_MIRROR = ~WSET; /* Disable interrupts using mirror */
	if (*INTENAR == 0)
	{
		/* Interrupts have been disabled. Maybe mirror of INTENA. */
		*INTENA_MIRROR = WSET | TBE; /* Enable TBE interrupt */
		if (*INTENAR != 0)
		{
			/* Interrupt was enabled. This is an INTENA mirror. */
			/* Restore interrupts */
			*INTENA = ~WSET;
			*INTENA = WSET | save_intena;

			/* So this is not a Gayle */
			return;
		}
	}

	/* On A300, we must clear the Fat Gary Timeout register
	 * to avoid reading a bogus 0x80 Gayle ID */
	FORCE_READ(FAT_GARY_TIMEOUT);

	gayle_id = 0;
	GAYLE_ID = 0; /* Reset ID register */
	for (i = 0; i < 8; i++)
	{
		gayle_id <<= 1;
		gayle_id |= (GAYLE_ID & 0x80) ? 1 : 0;
	}

	if (gayle_id == 0 || gayle_id == 0xff)
	{
		/* Bogus ID */
		return;
	}

	/* Gayle ID should be 0xd0 on ECS, or 0xd1 on AGA */
#ifdef DEBUG
	printf("gayle_id = 0x%02x\n", gayle_id);
#endif
	has_gayle = 1;
}

PRIVATE int readRDSK(dev)
short dev;
{
	long *b;
	long ret;
	int i;
#ifdef DEBUG
	printf("readRDSK(%d)\n",dev);
#endif
	for (i = 0; i < 16; i++) {
		b=(long*)xfer_buf;
		/*b=block;*/
		if (HDrdwr(dev, b, SECTOR_SIZE, (long)SECTOR_SIZE*i, DISK_READ))
			return(0);	/* Error: no RDSK found */
		if (b[0] == ID_RDSK)
			return(1);
	}
	return(0);
}

PRIVATE void readPART(dev, nr_spcyl, pb, pp)
short dev, nr_spcyl;
long pb;
struct part *pp;
{
	int i=0;
	static header=0;
	long *b;
#ifdef DEBUG
	printf("readPART(%d,%d,%ld,0x%8lx\n",dev,nr_spcyl,pb,pp);
#endif
		
	b=(long *)xfer_buf;
	/*b=block;*/
	while (pb !=-1)
	{
		i++; pp++; /* skip first entry (whole disk) */
		
		if (HDrdwr(dev, b, SECTOR_SIZE, (long)SECTOR_SIZE*pb, DISK_READ)!=OK)
		{
			printf("error while reading PART list on device %d\n",dev);
			return;
		}

#ifdef DEBUG
		printf("0x%8lx\n",b[0]);
#endif
		if (b[0] != ID_PART)
		{
			printf("corrupt PART list on device %d\n", dev);
			return;
		}
		pp->p_start=b[41]*nr_spcyl*SECTOR_SIZE;
		pp->p_size=(b[42]+1)*nr_spcyl*SECTOR_SIZE - pp->p_start;
		if (header==0)
		{
			header=1;
			printf("\nMinor  Size (MB)  Start  Name\n");
		}
		printf("%4d%8ld   %8ld   %s\n", dev + i, pp->p_size / 1048576,pp->p_start,BtoCstr(b+9));
		pb=b[4];
	}
}

init_dev(lun)
short lun;
{
	register struct part *pp;
	register int dev,i,nr_spcyl;
	long *b;

	dev = (lun<<3);
	pp = &part[dev];
	for (i=0; i<8; i++, pp++)
		pp->p_size=0;

	pp = &part[dev];
	pp->p_size = -1;

	if (readRDSK(dev)==1)
	{
		b=(long*)xfer_buf;
		/*b=block;*/
		nr_spcyl=b[17]*b[18];
		pp->p_size=b[16]*nr_spcyl*SECTOR_SIZE;
		readPART(dev,nr_spcyl,b[7],pp);
	}
}

PRIVATE int HDrdwr(minor, addr, nbytes, offset, cmd)
short minor;
char *addr;
u_short nbytes;
u_long offset;
int cmd;
{
  int lun;
  long ret;
  char *save_addr = addr;

  addr = xfer_buf;		/* be sure to use FASTMEM */
  if (cmd == DISK_WRITE) phys_copy(save_addr, addr, (long) nbytes);

  offset += part[minor].p_start;
  lun = (minor >> 3) & 1; /* 0 or 1 */

if (cmd==DISK_READ)
	ret=reg_pio_data_in_lba28( lun, CMD_READ_SECTORS,
                           0, nbytes / SECTOR_SIZE,
                           (long) offset / SECTOR_SIZE,
                           addr,
                           (long) nbytes / SECTOR_SIZE, 0 );
else
	ret=reg_pio_data_out_lba28( lun, CMD_WRITE_SECTORS,
                            0, nbytes / SECTOR_SIZE,
                            (long) offset / SECTOR_SIZE,
                            addr,
                            (long) nbytes / SECTOR_SIZE, 0 );

  if (cmd == DISK_READ) phys_copy(addr, save_addr, (long) nbytes);

#ifdef DEBUG
	printf("LUN:%d, RET:%ld, min: %d, addr: 0x%08lx, nbyts: %d, offs: %ld, cmd: %d)\n",lun,ret,minor,addr,nbytes,offset,cmd);
#endif

  if (ret!=OK){
	return(ERROR);
  }

	return(OK);
}

PRIVATE int do_rdwt(mp)
register message *mp;
{
  struct proc *rp;
  int r, rw, minor;
  unsigned int count;
  long offset;
  phys_bytes address;

  if (!has_gayle) {				/* No HD connected */
	static int err_shown = 0;

	if (!err_shown) {
		err_shown = 1;
		printf("Couldn't find ATA interface!\n");
	}
	return(EIO);
  }
  rw = mp->m_type;
  minor = mp->DEVICE;
  if (minor < 0 || minor > 15)
	return(EIO);
  if (part[minor].p_size == 0)		/* Illegal partition. */
	return(EIO);

  offset = mp->POSITION;		/* Offset and count must be a */
  if ((offset % SECTOR_SIZE) != 0)	/* multiple of SECTOR_SIZE. */
	return(EINVAL);

  count = mp->COUNT;
  if ((count % SECTOR_SIZE) != 0)
	return(EINVAL);

  if (count > MaxTransfer)
	return(EINVAL);

  if (count == 0)
	return(0);

  if (offset + count > part[minor].p_size) {
	if (offset >= part[minor].p_size)
		return(0);
	count = part[minor].p_size - offset;
  }

  rp = proc_addr(mp->PROC_NR);

  address = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) count);
  if (address == 0)
	return(EINVAL);

  rp->p_physio = 1;			/* Disable (un)shadowing. */

  r = HDrdwr(minor, address, count, offset, mp->m_type);

  rp->p_physio = 0;			/* Enable (un)shadowing. */
  if (r != OK)
	return(EIO);
  return(count);			/* Return number of bytes xferred. */
}

/*===========================================================================*
 *				winchester_task				     *
 *===========================================================================*/
PUBLIC void winchester_task()
{
	/* Main program of the ATA disk driver task. */
	int r, caller, proc_nr, dev, ndev;
	long ret;

	/* disable int */
	LOCK;
	
	detect_gayle();
	
	/* enable int */
	UNLOCK;
	
	/* Enable IDE int */
	*GAYLE_INT_1200 |= GAYLE_INT_IDE;
	
	/* First initialize the controller */
	ndev=reg_config();

	/* Allocate fast memory for the data buffer */
	xfer_buf=(char*)AllocMem((long)MaxTransfer,M_FAST32);
	
	for (dev=0;dev<ndev;dev++)
	{
		reg_reset(dev);
		init_dev(dev);
	}
	
#ifdef DEBUG
	printf("MaxTransfer: %u\n",MaxTransfer);
#endif


	/* Here is the main loop of the disk task.  It waits for a message, carries
	 * it out, and sends a reply.
	 */


	while (TRUE) {
		/* First wait for a request to read or write a disk block. */
		receive(ANY, &w_mess);	/* get a request to do some work */
		if (w_mess.m_source < 0) {
			panic("ATA task got message from %d\n",w_mess.m_source);
			/*continue;*/
		}
		caller = w_mess.m_source;
		proc_nr = w_mess.PROC_NR;

		/* Now carry out the work. */
		switch(w_mess.m_type) {
/* NOTE: these macros are changed in Minix 1.6.x */
			case DISK_READ:
			case DISK_WRITE:	r = do_rdwt(&w_mess);	break;
			case SCATTERED_IO:	r = do_vrdwt(&w_mess, do_rdwt); break;
			default:		r = EINVAL;			  
			break;
		}

		/* Finally, prepare and send the reply message. */
		w_mess.m_type = TASK_REPLY;
		w_mess.REP_PROC_NR = proc_nr;

		w_mess.REP_STATUS = r;	/* # of bytes transferred or error code */
		send(caller, &w_mess);	/* send reply to caller */
	}
}

PRIVATE char *BtoCstr(p)		/* Convert BCPL strings to C string */
char *p;
{
  p[(*p)+1] = 0;
  return(p+1);
}

/*===========================================================================*
 *				SYSTEM_READ_TIMER				     	*
 * from the file fs/utility.c but no super block update				*
 *===========================================================================*/
time_t SYSTEM_READ_TIMER()
{
/* This routine returns the time in seconds since 1.1.1970.  MINIX is an
 * astrophysically naive system that assumes the earth rotates at a constant
 * rate and that such things as leap seconds do not exist.
 */

  register int k;

  clock_mess.m_type = GET_TIME;
  if ( (k = sendrec(CLOCK, &clock_mess)) != OK) panic("SYSTEM_READ_TIMER err", k);

  return (time_t) clock_mess.NEW_TIME;
}

int SYSTEM_WAIT_INTR_OR_TIMEOUT() 
{
	int ret;

	ret=receive(HARDWARE, &dummy_mess) ? 1 : 0;
	int_ata_status = pio_inbyte(CB_STAT);
	return ret;
}

/* NOTE: replace the call to a590_int in the kernel/main.c file with this function's name
*/
PUBLIC ata_int()
{

	/* Test IRQ */
	if ((*GAYLE_IRQ_1200) & GAYLE_IRQ_IDE)
	{
		/* Handle spurious interrupt */
		int_ata_status=pio_inbyte( CB_STAT );
	
		/* ACK = Clear A600/A1200 IDE interrupt. (Stupid Gayle hardware) */ 
		/* Ideally this should be done with interrupts disabled */	
		*GAYLE_IRQ_1200 = 0x7c | (*GAYLE_IRQ_1200 & 3);
	
		if (int_ata_status & CB_STAT_BSY) /* ATA interrupt but BUSY flag set!? */
			return;

		interrupt(WINCHESTER);
	}
	return;
}

